import base64
import os
import urllib
import warnings
from urllib.parse import urlparse

from chromepdf.conf import parse_settings
from chromepdf.pdfconf import clean_pdf_kwargs
from chromepdf.webdrivers import (
    _get_chromesession_temp_dir, devtool_command, download_chromedriver_version, get_chrome_version,
    get_chrome_webdriver)


class ChromePdfMaker:
    """A class used to expedite PDF creation and storing of settings state."""

    def __init__(self, **kwargs):

        # load settings, combining those from **kwargs as well as the Django settings.
        settings = parse_settings(**kwargs)
        self._chrome_path = settings['chrome_path']
        self._chromedriver_path = settings['chromedriver_path']
        self._chromedriver_downloads = settings['chromedriver_downloads']
        self._chromesession_temp_dir = _get_chromesession_temp_dir()

        os.makedirs(self._chromesession_temp_dir, exist_ok=True)

        # download chromedriver if we have chrome, and downloads are enabled
        if self._chrome_path is not None and self._chromedriver_path is None and self._chromedriver_downloads:
            chrome_version = get_chrome_version(self._chrome_path, as_tuple=False)
            self._chromedriver_path = download_chromedriver_version(chrome_version)

        self._webdriver_kwargs = {
            'chrome_args': settings['chrome_args'],
            'chrome_path': self._chrome_path,
            'chromedriver_path': self._chromedriver_path,
            '_chromesession_temp_dir': self._chromesession_temp_dir,
        }

    def _clean_pdf_kwargs(self, pdf_kwargs):
        """A wrapper around clean_pdf_kwargs() that handles None as well."""

        pdf_kwargs = {} if pdf_kwargs is None else pdf_kwargs
        pdf_kwargs = clean_pdf_kwargs(**pdf_kwargs)
        return pdf_kwargs

    def generate_pdf(self, html, pdf_kwargs=None):
        """Generate a PDF file from an html string and return the PDF as a bytes object."""

        pdf_kwargs = self._clean_pdf_kwargs(pdf_kwargs)

        with get_chrome_webdriver(**self._webdriver_kwargs) as driver:

            # we could put the html here. but data urls in Chrome are limited to 2MB.
            dataurl = "data:text/html;charset=utf-8," + urllib.parse.quote('')
            driver.get(dataurl)

            # append our html. theoretically no length limit.
            html = html.replace('`', r'\`')  # escape the backtick used to indicate a multiline string in javascript
            # we do NOT need to escape any other chars (quotes, etc), including unicode
            driver.execute_script("document.write(`{}`)".format(html))

            result = devtool_command(driver, "Page.printToPDF", pdf_kwargs)

        outbytes = base64.b64decode(result['data'])
        return outbytes

    def generate_pdf_url(self, url, pdf_kwargs=None):
        """Generate a PDF file from a url (such as a file:/// url) and return the PDF as a bytes object."""

        warnings.warn("ChromePdfMaker.generate_pdf_url() is deprecated, use generate_pdf() instead.", DeprecationWarning)

        # throw an early exception if we receive a string that Chrome would return a 400 error (Bad Request) if given.
        parseresult = urlparse(url)
        if not parseresult.scheme:
            raise ValueError('generate_pdf_url() requires a valid URI, beginning with file:/// or https:// or similar. '
                             'You can use: import pathlib; pathlib.Path(absolute_path).as_uri() to '
                             'convert an absolute path into such a file URI.')

        pdf_kwargs = self._clean_pdf_kwargs(pdf_kwargs)

        with get_chrome_webdriver(**self._webdriver_kwargs) as driver:
            driver.get(url)
            result = devtool_command(driver, "Page.printToPDF", pdf_kwargs)

        outbytes = base64.b64decode(result['data'])
        return outbytes



import string
from decimal import Decimal


# For more information, see: https://www.w3.org/TR/CSS2/syndata.html#length-units
# NOTE: All keys should be two chars long. All values should be integers or floats.
UNITS_PER_INCH = {
    'px': 96,  # these are "Reference Pixels" (defined by CSS standard as 96 DPI (.75 px/pt, and 72 pt/in = 96).
    'cm': 2.54,
    'mm': 25.4,
    'in': 1,
}
UNIT_STR_LENGTH = 2  # our parser expects all unit types to be two characters long ("in", "cm", etc)

# All values MUST be provided as integers or floats, and given in inches.
# These values will be passed to Page.PrintToPDF, which only accepts inches.
# These sizes are the same ones as supported  by PhantomJS, plus ledger
# A good reference for these sizes (used for printers): https://doc.qt.io/archives/qt-5.10/qpagesize.html
PAPER_FORMATS = {
    'ledger': {'width': 17, 'height': 11},
    'legal': {'width': 8.5, 'height': 14},
    'letter': {'width': 8.5, 'height': 11},
    'tabloid': {'width': 11, 'height': 17},
    'a3': {'width': 11.69, 'height': 16.54},
    'a4': {'width': 8.26, 'height': 11.69},
    'a5': {'width': 5.83, 'height': 8.26},
}


def convert_to_inches(size):
    """
    Take a case-insensitive string consisting of a CSS length and return a float or int of that value in inches.
    If an int or float is passed in, then it is assumed to be an int and is returned as-is.
    If None is passed, return None (presumed that we're trying to "disable" the setting).
    """

    if size is None:
        return None
    elif isinstance(size, (int, float, Decimal)):  # inches
        if isinstance(size, Decimal):
            return float(size)  # we must return the value in a json-serializable format, which Decimal is not.
        return size  # int and float are json-serializable
    elif isinstance(size, str) and len(size) > UNIT_STR_LENGTH:
        format_num = size[:-UNIT_STR_LENGTH]  # EG "2.5in" => "2.5"
        format_str = size[-UNIT_STR_LENGTH:].lower()  # EG "2.5in" => "in"
        # invalid unit type?
        if format_str not in UNITS_PER_INCH:
            raise ValueError(f'generate_pdf() cannot understand a length value of: "{size}"')
        # if this contains any whitespace, raise value error (not valid CSS)
        if any(c in string.whitespace for c in format_num):
            raise ValueError(f'generate_pdf() cannot understand a length value of: "{size}"')
        try:
            value = float(format_num)

            return value / UNITS_PER_INCH[format_str]  # convert to inches
        except ValueError as ex:
            raise ValueError(f'generate_pdf() cannot understand a length value of: "{size}"') from ex
    else:
        raise ValueError(f'generate_pdf() cannot understand a length value of: "{size}"')


def convert_to_unit(size, unit):
    """
    Take any int/float size value, and return an int/float of that value in the specified unit type.
    convert_to_unit('1in', 'cm') => 2.54
    """
    if unit not in UNITS_PER_INCH:
        raise ValueError(f'Invalid unit type: {unit}')
    return convert_to_inches(size) * UNITS_PER_INCH[unit]



import warnings

from chromepdf.maker import ChromePdfMaker


def generate_pdf(html, pdf_kwargs=None, **kwargs):
    """
    Return the bytes of a PDF file that is generated from the HTML and pdf_kwargs passed in.

    html: A string
    pdf_kwargs: A dict containing any of the arguments accepted by Chrome's Page.printToPDF API.
    See the clean_pdf_kwargs() docstring below for a list of valid options.
    **kwargs: Lowercased settings such as chrome_path, and chromedriver_path. See conf.py's DEFAULT_SETTINGS dict.
    """

    return ChromePdfMaker(**kwargs).generate_pdf(html, pdf_kwargs)


def generate_pdf_url(url, pdf_kwargs=None, **kwargs):
    """
    NOTE: This function is DEPRECATED due to security concerns due to its ability to pull in any files on the server.
    Users should switch to using generate_pdf() as soon as possible.

    Return the bytes of a PDF file that is generated from the HTML and pdf_kwargs passed in.

    url: A url to load (may be local, too. EG: "file:///C:/Users/MyName/../../myfile.html")
    pdf_kwargs: A dict containing any of the arguments accepted by Chrome's Page.printToPDF API.
    See the clean_pdf_kwargs() docstring below for a list of valid options.
    **kwargs: Lowercased settings such as chrome_path, and chromedriver_path. See conf.py's DEFAULT_SETTINGS dict.

    If you have a local path that needs to be converted to a URI, you can import pathlib and call html_uri = pathlib.Path(mypath).as_uri()
    Sample use:

    from chromepdf import generate_pdf_url

    html_uri = "file:///C:/Users/myuser/Desktop/myfile.html"
    pdf_kwargs = {
        'paperFormat': 'A4',
        'marginTop': '2.5cm',
        'marginLeft': '2cm',
        'marginRight': '2cm',
        'marginBottom': '3.5cm',
        'displayHeaderFooter': True,
        'headerTemplate': '',
        'footerTemplate': '''
            <div style="font-size: 12px; width: 100%; padding: 0; padding-left: 2cm; padding-bottom: 1cm; margin: 0; ">
                Page <span class="pageNumber"></span> of <span class="totalPages"></span>
            </div>
        ''',
    }
    pdfbytes = generate_pdf_url(html_uri, pdf_kwargs)

    with open("myfile.pdf", 'wb') as file:
        file.write(pdfbytes)


    """

    warnings.warn("generate_pdf_url() is deprecated, use generate_pdf() instead.", DeprecationWarning)

    return ChromePdfMaker(**kwargs).generate_pdf_url(url, pdf_kwargs)



__version__ = '1.5.0'

from chromepdf.exceptions import ChromePdfException
from chromepdf.maker import ChromePdfMaker
from chromepdf.shortcuts import generate_pdf, generate_pdf_url


__all__ = ['__version__',
           'generate_pdf', 'generate_pdf_url', 'ChromePdfException', 'ChromePdfMaker']



DEFAULT_SETTINGS = {
    'CHROME_PATH': None,
    'CHROMEDRIVER_PATH': None,
    'CHROMEDRIVER_DOWNLOADS': True,
    # also, PDF_KWARGS, but it's handled differently
    'CHROME_ARGS': [],
}


def get_chromepdf_settings_dict():
    """
    Return a Django's settings.CHROMEPDF dict. Return empty dict if not found or Django not installed.
    For our sanity, this should be the ONLY place within the chromepdf app that we import from Django.
    This way, the library should work even if Django is not installed (except with its settings ignored).
    """
    try:
        from django.conf import settings
        return getattr(settings, 'CHROMEPDF', {})
    except Exception:
        # catches ImportError and ModuleNotFoundError, and potentially Django's ImproperlyConfigured.
        # But we can't explitly name that here since Django might not be installed. So, use a general Exception.
        return {}


def parse_settings(**overrides):
    """
    Return a dict of lowercased DEFAULT_SETTINGS based on combination of defaults, Django settings, and overrides.
    Priority: keyword overrides > Django settings > DEFAULT_SETTINGS
    """

    output = {}
    chromepdf_settings = get_chromepdf_settings_dict()

    # iterate over expected setting names
    for k, defaultval in DEFAULT_SETTINGS.items():
        k_lower = k.lower()
        if k_lower in overrides:
            output[k_lower] = overrides[k_lower]  # get kwarg
        else:
            output[k_lower] = chromepdf_settings.get(k, defaultval)  # get Django setting, OR default value

        # convert falsey values to more appropriate ones.
        if k == 'CHROMEDRIVER_DOWNLOADS':  # boolean settings
            if output[k_lower] is None:
                output[k_lower] = False
        elif k == 'CHROME_ARGS':
            if output[k_lower] is None:
                output[k_lower] = []
            elif isinstance(output[k_lower], str):
                # Prevent passing individual characters in str to Chrome as args. Silent, and unexpected behavior.
                raise TypeError('The chrome_args/CHROME_ARGS parameter/setting must be an iterable of strings.')
        else:  # path settings
            if output[k_lower] == '':
                output[k_lower] = None

    return output



"""
Command-line entry point for running ChromePDF.
To execute, run:
> python -m chromepdf generate-pdf [args] [kwargs]
"""


def main():
    """
    To make unit testing chromepdf.__main__ easier, put everything into a function
    Except the single line executed by __name__ == '__main__', which is hard if not impossible to get at.
    """
    import sys

    from .run import chromepdf_run
    chromepdf_run(sys.argv[1:])


if __name__ == '__main__':
    main()




def chromepdf_run(args):
    """
    A method of generating PDF files from the command line. To execute, run:
    > python -m chromepdf generate-pdf [args] [kwargs]
    """

    parser = _get_parser()
    namespace = parser.parse_args(args)

    if namespace.command == 'generate-pdf':
        _command_generate_pdf(parser, namespace)
    else:
        parser.print_help()
        # 'Unix programs generally use 2 for command line syntax errors and 1 for all other kind of errors.'
        # If execution failed to run a reasonable task with command, return an error response.
        exit(2)


def _get_parser():
    """Return an ArgumentParser that handles chromepdf commands."""

    import argparse
    parser = argparse.ArgumentParser()

    subparsers = parser.add_subparsers(help='You may call the following commands:', dest='command')

    # For now, we only have one sub-command: generate-pdf.
    # Other sub-commands may be added in the future
    genpdf_parser = subparsers.add_parser('generate-pdf', help='Generate a PDF file. Followed by one or two args: The part to the input HTML file, and path to the output PDF file. EG: "generate-pdf path/to/file.html path/to/file.pdf"')
    genpdf_parser.add_argument('paths', nargs='*')
    genpdf_parser.add_argument("--pdf-kwargs-json", help="Path to a JSON file whose contents can decode to a pdf_kwargs dict.")
    genpdf_parser.add_argument("--chrome-path", help="Path to Chrome executable")
    genpdf_parser.add_argument("--chromedriver-path", help="Path to Chrome executable")
    genpdf_parser.add_argument("--chromedriver-downloads", type=int, choices=(0, 1), help='1 or 0, to indicate whether to use Chromedriver downloads or not.')
    genpdf_parser.add_argument("--chrome-args", help='A string of all arguments to pass to Chrome, separated by spaces.')

    return parser


def _command_generate_pdf(parser, namespace):
    """Call the generate_pdf() function using command-line arguments."""

    import os

    if namespace.paths is None or len(namespace.paths) == 0:
        parser.error('generate-pdf: requires one or two path arguments for an infile and optional outfile.')

    if len(namespace.paths) == 1:
        inpath = namespace.paths[0]
        outpath = os.path.splitext(inpath)[0] + '.pdf'  # replace extension with pdf. OR append if has none.
    elif len(namespace.paths) == 2:
        inpath, outpath = namespace.paths
    else:
        parser.error('generate-pdf: requires one or two path arguments for an infile and optional outfile.')

    if not os.path.exists(inpath):
        parser.error(f'generate-pdf: could not find input html file: "{inpath}"')

    with open(inpath, 'r', encoding='utf8') as f:
        html_str = f.read()

    kwargs = {}
    if namespace.chrome_path is not None:
        kwargs['chrome_path'] = namespace.chrome_path
    if namespace.chromedriver_path is not None:
        kwargs['chromedriver_path'] = namespace.chromedriver_path
    if namespace.chromedriver_downloads is not None:
        kwargs['chromedriver_downloads'] = bool(namespace.chromedriver_downloads)
    if namespace.chrome_args is not None:
        kwargs['chrome_args'] = namespace.chrome_args.strip().split()

    pdf_kwargs = None
    if namespace.pdf_kwargs_json is not None:
        if not os.path.exists(namespace.pdf_kwargs_json):
            parser.error(f'generate-pdf: could not find input pdf-kwargs-json file: "{namespace.pdf_kwargs_json}"')
        try:
            import json
            with open(namespace.pdf_kwargs_json, 'r', encoding='utf8') as f:
                data = f.read()
                pdf_kwargs = json.loads(data)
            if not isinstance(pdf_kwargs, dict):
                parser.error('--pdf-kwargs-json: must be a path to a file containing a JSON dict {} of key-value pairs. The JSON in this file is a different data type.')
        except json.JSONDecodeError:
            parser.error('--pdf-kwargs-json: must be a path to a file containing a JSON dict {} of key-value pairs. The JSON in this file is not valid JSON.')

    from .shortcuts import generate_pdf
    pdf_bytes = generate_pdf(html_str, pdf_kwargs, **kwargs)
    outpath_dir = os.path.dirname(outpath)
    if outpath_dir:  # makedirs() will fail if we try passing the current dir via "", so don't.
        os.makedirs(outpath_dir, exist_ok=True)
    with open(outpath, 'wb') as f:
        f.write(pdf_bytes)



from chromepdf.conf import get_chromepdf_settings_dict
from chromepdf.sizes import PAPER_FORMATS, convert_to_inches


# Specified in printToPDF API - https://chromedevtools.github.io/devtools-protocol/1-3/Page/#method-printToPDF
# These are the "TRUE" defaults.
# If no overrides are provided in Django's settings.CHROMEPDF['PDF_KWARGS'], and no pdf_kwargs are passed to render functions,
# Then these are the parameters that would be used and passed to Chrome.
# You do NOT need to call convert_to_inches() when overriding these values yourself.
DEFAULT_PDF_KWARGS = dict(
    landscape=False,
    displayHeaderFooter=False,
    printBackground=False,
    scale=1,
    paperWidth=convert_to_inches('8.5in'),
    paperHeight=convert_to_inches('11in'),
    marginTop=convert_to_inches('1cm'),
    marginLeft=convert_to_inches('1cm'),
    marginRight=convert_to_inches('1cm'),
    marginBottom=convert_to_inches('1cm'),
    pageRanges='',
    ignoreInvalidPageRanges=False,
    headerTemplate='',
    footerTemplate='',
)


def get_default_pdf_kwargs():
    """
    Return the default pdf_kwargs used for rendering a PDF in Chrome.
    Use the values specified in Django settings.CHROMEPDF['PDF_KWARGS'] if they exist.
    Otherwise, fallback to the `DEFAULT_PDF_KWARGS` above.
    """

    defaults = {}
    defaults.update(DEFAULT_PDF_KWARGS)  # make sure we're working on a copy of the settings.

    settings_pdf_kwargs = get_chromepdf_settings_dict().get('PDF_KWARGS', {})
    if settings_pdf_kwargs:
        # clean the overrides using the "true" defaults as defaults.
        # When we call the generate_pdf() functions, the resulting combined "defaults" will be treated as defaults.
        # This way, when we fall back on the defaults for any not-provided arguments,
        # we can be sure they are in a consistent format.
        overrides = clean_pdf_kwargs(_defaults=defaults, **settings_pdf_kwargs)
        defaults.update(overrides)

    return defaults


def clean_pdf_kwargs(**options):
    """
    Clean the pdf_kwargs into a format that Page.printToPDF accepts.
    Our defaults should match the default arguments of the Page.printToPDF API.
    For more information, see: https://chromedevtools.github.io/devtools-protocol/1-3/Page/#method-printToPDF

    :param scale: Scale of the PDF. Default 1.
    :param landscape: True to use landscape mode. Default False.

    :param displayHeaderFooter: True to display header and footer. Default False.
    :param headerTemplate: HTML containing the header for all pages. Default is an empty string.
        You may pass html tags with specific classes in order to insert values:
            * date: formatted print date
            * title: document title
            * url: document location
            * pageNumber: current page number
            * totalPages: total pages in the document
        For example, <span class="title"></span> would generate span containing the title.
    :param footerTemplate: HTML containing the footer for all pages. Default is an empty string.
        You may pass html tags with specific classes in order to insert values (same as above)
    :param printBackground: True to print background graphics. Default False.

    :param paperWidth: Width of the paper, in inches. Can also use some CSS string values, like "30cm". Default: 8.5
    :param paperHeight: Height of the paper, in inches. Can also use some CSS string values, like "30cm". Default: 11
    :param paperFormat: A string indicating a paper size format, such as "letter" or "A4". Case-insensitive.
        This will override paperWidth and paperHeight.
        Not part of Page.printToPDF API.  Provided for convenience.

    :param margin: Shortcut used to set all four margin values at once.
        Not part of Page.printToPDF API.  Provided for convenience.
    :param marginTop: Top margin. Default '1cm'
    :param marginBottom: Bottom margin. Default '1cm'.
    :param marginLeft: Left margin. Default '1cm'.
    :param marginRight: Right margin. Default '1cm'.

    :param pageRanges: String indicating page ranges to use. Example: '1-5, 8, 11-13'
    :param ignoreInvalidPageRanges: If True, will silently ignore invalid "pageRanges" values. Default False.

    :param _defaults: Internal-only. An optional dict of default parameter overrides that have already been 'cleaned'.

    :return: A dict containing an options dict that be used for Page.printToPDF
    """

    if '_defaults' in options:
        defaults = options.pop('_defaults')
    else:
        defaults = get_default_pdf_kwargs()

    try:
        scale = float(options.get('scale', defaults['scale']))  # can be int or float, or numeric string
    except ValueError as ex:  # passed a character string?
        raise TypeError('You must pass a numeric value for the "scale" of the PDF.') from ex

    landscape = bool(options.get('landscape', defaults['landscape']))

    displayHeaderFooter = bool(options.get('displayHeaderFooter', defaults['displayHeaderFooter']))
    headerTemplate = options.get('headerTemplate', defaults['headerTemplate'])
    footerTemplate = options.get('footerTemplate', defaults['footerTemplate'])
    printBackground = bool(options.get('printBackground', defaults['printBackground']))

    paperWidth = defaults['paperWidth']
    paperHeight = defaults['paperHeight']
    if 'paperFormat' in options:  # convenience option that's not part of Chrome's API
        if options['paperFormat'].lower() not in PAPER_FORMATS:
            raise ValueError(f'Unrecognized paper format: "{options["paperFormat"]}"')
        if 'paperWidth' in options or 'paperHeight' in options:
            raise ValueError('Cannot pass a paperFormat at the same time as a paperWidth/paperHeight.')

        paperFormat = PAPER_FORMATS.get(options.pop('paperFormat').lower())  # pop it so we can validate options below
        paperWidth = paperFormat['width']
        paperHeight = paperFormat['height']
    else:
        if 'paperWidth' in options:
            paperWidth = convert_to_inches(options['paperWidth'])
        if 'paperHeight' in options:
            paperHeight = convert_to_inches(options['paperHeight'])

    if paperWidth is None:
        raise TypeError('You must set a paperWidth for this PDF.')
    if paperHeight is None:
        raise TypeError('You must set a paperHeight for this PDF.')

    marginTop = defaults['marginTop']
    marginBottom = defaults['marginBottom']
    marginLeft = defaults['marginLeft']
    marginRight = defaults['marginRight']

    # margin affects all four sides. margin in options will override default side-specific defaults.
    if 'margin' in options:
        margin = convert_to_inches(options.pop('margin'))
        marginTop = marginBottom = marginLeft = marginRight = margin

    # margin overrides for specific sides
    marginTop = convert_to_inches(options.get('marginTop', marginTop))
    marginBottom = convert_to_inches(options.get('marginBottom', marginBottom))
    marginLeft = convert_to_inches(options.get('marginLeft', marginLeft))
    marginRight = convert_to_inches(options.get('marginRight', marginRight))

    if marginTop is None:
        raise TypeError('You cannot set marginTop to None')
    if marginBottom is None:
        raise TypeError('You cannot set marginBottom to None')
    if marginLeft is None:
        raise TypeError('You cannot set marginLeft to None')
    if marginRight is None:
        raise TypeError('You cannot set marginRight to None')

    pageRanges = str(options.get('pageRanges', defaults['pageRanges']))
    ignoreInvalidPageRanges = bool(options.get('ignoreInvalidPageRanges', defaults['ignoreInvalidPageRanges']))
    # transferMode: not applicable.

#     preferCSSPageSize = options.get('preferCSSPageSize','')

    # the actual dict that we will pass to Chrome (convenience kwargs like margin and paperFormat are removed)
    parameters = dict(
        scale=scale,
        landscape=landscape,
        #         preferCSSPageSize=preferCSSPageSize, # too new, causes error

        displayHeaderFooter=displayHeaderFooter,
        headerTemplate=headerTemplate,
        footerTemplate=footerTemplate,
        printBackground=printBackground,

        paperWidth=paperWidth,
        paperHeight=paperHeight,

        marginTop=marginTop,
        marginBottom=marginBottom,
        marginLeft=marginLeft,
        marginRight=marginRight,

        pageRanges=pageRanges,
        ignoreInvalidPageRanges=ignoreInvalidPageRanges,
    )

    # check for bad options
    parameters_keys = set(parameters.keys())
    options_keys = set(options.keys())
    if not options_keys.issubset(parameters_keys):
        unrecognized_keys = ', '.join(sorted(options_keys - parameters_keys))
        raise ValueError(f'Unrecognized pdf_kwargs passed to generate_pdf(): {unrecognized_keys}')

    return parameters



import getpass
import io
import json
import os
import platform
import shutil
import subprocess
import warnings
import zipfile
from contextlib import contextmanager
from subprocess import PIPE
from urllib import request as urllib_request

import selenium
from selenium import webdriver
from selenium.webdriver.chrome.service import Service

from chromepdf.exceptions import ChromePdfException


_IS_SELENIUM_3 = selenium.__version__.split('.')[0] == '3'


def _version_to_tuple(version):
    """
    Shortcut function that converts version string to tuple, for allowing deprecated behavior.
    In ChromePDF 2.0, this should no longer be needed. Tuple support will no longer exist.
    """
    return tuple(int(i) for i in version.split('.'))


def _force_version_str(version):
    """
    Shortcut function that lets us accept tuple versions and force them into string versions.
    In ChromePDF 2.0, this should no longer be needed. Tuple support will no longer exist.
    """
    if isinstance(version, tuple):
        warnings.warn("Support for versions as tuples is now deprecated. Pass a string instead.", DeprecationWarning)
        return '.'.join(str(i) for i in version)
    return version


def _get_chrome_version_str(path):
    """
    Return a string containing the version number of the Chrome binary exe, EG for Chrome 85: '85.0.4183.121'
    raise ChromePdfException otherwise (EG if path not found)
    """

    is_windows = (platform.system() == 'Windows')

    try:
        if is_windows:
            cmd = f'(Get-Item "{path}").VersionInfo'
            proc = subprocess.run(['powershell', cmd], check=True, stdout=PIPE, stderr=PIPE)
            # NOTE: "chrome.exe --version" does NOT work on windows. This is one workaround.
            # https://bugs.chromium.org/p/chromium/issues/detail?id=158372
            #
            # this will output a table like so. In this case, we want to grab the "85.0.4183.121"
            # ProductVersion   FileVersion      FileName
            # --------------   -----------      --------
            # 85.0.4183.121    85.0.4183.121    C:\Program Files (x86)\Google\Chrome\Application\chrome.exe
            lines = [l.strip() for l in proc.stdout.decode('utf8').split('\n') if l.strip()]
            for l in lines:
                if l[0].isdigit():
                    return l.split()[0]

            # if no lines containing version numbers were output
            raise ChromePdfException(f'Could not determine version of Chrome executable located at: "{path}"')

        else:  # linux, mac can both just use "--version"
            proc = subprocess.run([path, '--version'], check=True, stdout=PIPE, stderr=PIPE)
            version_stdout = proc.stdout.decode('utf8').strip()  # returns, eg, "Google Chrome 85.0.4183.121"
            return [i for i in version_stdout.split() if i[0].isdigit()][0]

    except Exception as ex:
        # FileNotFoundError, CalledProcessError, IndexError, etc
        if os.path.exists(path):
            exception_msg = f'Could not determine version of Chrome located at: "{path}"'
        else:
            exception_msg = f'Tried to determine version of Chrome located at: "{path}", but no executable exists at that location.'
        raise ChromePdfException(exception_msg) from ex


def get_chrome_version(path, as_tuple=True):
    """
    Return a 4-tuple containing the version number of the Chrome binary exe, EG for Chrome 85: (85,0,4183,121)
    Return a string instead if as_tuple=True is passed.
    raise ChromePdfException otherwise (EG if path not found)
    """

    version = _get_chrome_version_str(path)
    if as_tuple:
        warnings.warn("get_chrome_version() support for returning tuples is deprecated. Pass as_tuple=False instead. In ChromePDF 2.0, this function will return string values always.", DeprecationWarning)
        return _version_to_tuple(version)
    return version


def _get_chromedriver_environment_path():
    """
    Return full path of chromedriver on PATH, if it exists. Otherwise, return None.
    This can be used to determine if Selenium can find the chromedriver without having to specify its location.
    """

    is_windows = (platform.system() == 'Windows')
    filename = 'chromedriver.exe' if is_windows else 'chromedriver'
    return shutil.which(filename)


def _get_chromedriver_download_path(version):
    """Return a path to put/find a chromedriver file, if ChromePDF should/did download it."""

    version = _force_version_str(version)
    is_windows = (platform.system() == 'Windows')
    chromedrivers_dir = os.path.join(os.path.dirname(__file__), 'chromedrivers')
    chromedriver_path = os.path.join(chromedrivers_dir, f'chromedriver_{version}')
    if is_windows:  # windows requires an extension or it won't run.
        chromedriver_path += '.exe'
    else:  # linux, mac = no extension
        pass
    return chromedriver_path


def _fetch_chromedriver_version_for_chrome_version(version):
    """
    Fetch the chromedriver version needed for the given Chrome version by querying the official website.
    Returns a version string such as "85.0.4183.87"
    """

    version = _force_version_str(version)

    # Google's API for the latest release takes only the first 3 parts of the version
    version_first3parts = version.rsplit('.', maxsplit=1)[0]  # EG, "85.0.4183"

    # This url returns a 4-part version string of the latest compatible chromedriver for your Chrome version.
    # This might be DIFFERENT than the version of your Chrome executable.
    url = f'https://chromedriver.storage.googleapis.com/LATEST_RELEASE_{version_first3parts}'
    with urllib_request.urlopen(url) as f:
        contents = f.read()
    chromedriver_version = contents.decode('utf8')  # EG "85.0.4183.87"
    return chromedriver_version


def _get_chromedriver_zip_url(chromedriver_version):
    """
    Get the chromedriver zip download url for our particular OS+Processor.
    The possible urls are taken from the chromedriver release files list, here:
    https://chromedriver.chromium.org/downloads
    """

    is_windows = (platform.system() == 'Windows')
    is_mac = (platform.system() == 'Darwin')
    is_mac_m1 = (is_mac and platform.processor() == 'arm')

    if is_windows:
        os_plus_numbits = 'win32'
    elif is_mac_m1:
        os_plus_numbits = 'mac64_m1'
    elif is_mac:
        os_plus_numbits = 'mac64'
    else:
        os_plus_numbits = 'linux64'
    filename = f'chromedriver_{os_plus_numbits}.zip'

    return f'https://chromedriver.storage.googleapis.com/{chromedriver_version}/{filename}'


def _fetch_chromedriver_zip_bytes(chromedriver_version):
    """
    Return the bytes of the chromedriver zip file for the given chromedriver version, for our OS.
    """

    url = _get_chromedriver_zip_url(chromedriver_version)
    with urllib_request.urlopen(url) as f:
        zip_bytes = f.read()
    return zip_bytes


def download_chromedriver_version(version, force=False):
    """
    Download a chromedriver executable for the Chrome version specified, if not already downloaded or force=True.
    Return the path of the existing (or newly downloaded) chromedriver executable.

    See https://chromedriver.chromium.org/downloads/version-selection
    for download url api

    Arguments:
    * version: A version string as returned by get_chrome_version(), such as: '85.0.4183.121'
    * force: If True, will force a download, even if a driver for that version is already saved.
    """

    version = _force_version_str(version)

    # Return the existing path if it exists and we're not forcing a new download.
    chromedriver_download_path = _get_chromedriver_download_path(version)
    if os.path.exists(chromedriver_download_path) and not force:
        return chromedriver_download_path

    # chromedrivers have their own version strings. fetch the one for our chrome version.
    chromedriver_version = _fetch_chromedriver_version_for_chrome_version(version)

    # Download the zip file containing our chromedriver
    zip_bytes = _fetch_chromedriver_zip_bytes(chromedriver_version)

    # Open the zip file, find the chromedriver, and save it to the specified path.
    with zipfile.ZipFile(io.BytesIO(zip_bytes), "r") as zf:
        for name in zf.namelist():
            if 'chromedriver' in name:
                with zf.open(name) as chromedriver_file:
                    with open(chromedriver_download_path, 'wb') as f:
                        f.write(chromedriver_file.read())
                        os.chmod(chromedriver_download_path, 0o764)  # grant execute permission
                        return chromedriver_download_path

    raise ChromePdfException('Failed to download the chromedriver file.')


@contextmanager
def get_chrome_webdriver(chrome_path, chromedriver_path, **kwargs):
    """
    Create and return a Chrome webdriver. Is a context manager, and will automatically close the driver. Usage:

    * chromedriver_path: Path to your chromedriver executable. If None, will try to find it on PATH via Selenium.
    * chrome_path: Path to your Chrome exe. If None, driver will try to find it automatically.
    kwarg-only:
    * chrome_args: List of options to pass to Chrome.

    with get_chrome_webdriver(...) as driver:
        # call commands...
    # driver is automatically closed
    """

    chrome_webdriver_kwargs = _get_chrome_webdriver_kwargs(chrome_path, chromedriver_path, **kwargs)

    # contextmanager.__enter__
    try:
        driver = webdriver.Chrome(**chrome_webdriver_kwargs)
    except Exception as ex:
        if chrome_path and not os.path.exists(chrome_path):
            raise ChromePdfException(f'Could not find a chrome_path path at: {chrome_path}') from ex
        elif chromedriver_path and not os.path.exists(chromedriver_path):
            raise ChromePdfException(f'Could not find a chromedriver_path at: {chromedriver_path}') from ex
        else:
            raise ex

    yield driver

    # contextmanager.__exit__
    driver.quit()  # quits the entire driver (driver.close() only closes the current window)


def _get_chrome_webdriver_kwargs(chrome_path, chromedriver_path, **kwargs):
    """Return the kwargs needed to pass to webdriver.Chrome(), given the CHROMEPDF settings."""

    # at one point "-disable-gpu" was required for headless Chrome. Keep it here just in case.
    # https://bugs.chromium.org/p/chromium/issues/detail?id=737678
    options = webdriver.ChromeOptions()
    options.add_argument("--headless")
    options.add_argument("--disable-gpu")

    options.add_argument("--log-level=3")  # silence logging

    # disables the creation of the crash-dumps-dir.
    # will work even if --crash-dumps-dir is provided.
    # keep both as fallbacks. either is preferable to the global default.
    options.add_argument('--disable-crash-reporter')

    # incognito mode lets us run chrome without creating and storing a user profile to disk.
    # this lets us generate PDFs in a thread- and process-safe way since they will not be
    # fighting over reading/writing the same files in the --user-data-dir
    # passing --incognito AND --user-data-dir= WILL cause the user-data-dir folder to be populated, so don't.
    options.add_argument("--incognito")

    temp_dir = kwargs.get('_chromesession_temp_dir')
    if temp_dir is not None:
        crash_dumps_dir = os.path.join(temp_dir, 'crash-dumps-dir')
        options.add_argument(f"--crash-dumps-dir={crash_dumps_dir}")

    # add extra chrome args
    chrome_args = kwargs.get('chrome_args', [])
    for argv in chrome_args:
        options.add_argument(argv)

    # silence the "DevTools started" message on windows
    # https://bugs.chromium.org/p/chromedriver/issues/detail?id=2907#c3
    options.add_experimental_option('excludeSwitches', ['enable-logging'])

    # In Selenium 4, we must tell the driver to ignore any os.environ['http_proxy'/'https_proxy'] values.
    # Calling this function preserves Selenium 3 behavior, which did not check them at all.
    # This function does not exist in Selenium 3, so we must check if it exists to preserve Selenium 3 compatibility.
    # https://github.com/SeleniumHQ/selenium/issues/8768
    if hasattr(options, 'ignore_local_proxy_environment_variables') and callable(options.ignore_local_proxy_environment_variables):
        options.ignore_local_proxy_environment_variables()

    if chrome_path is not None:
        options.binary_location = chrome_path  # Selenium API

    chrome_kwargs = {'options': options}
    if chromedriver_path is not None:
        if _IS_SELENIUM_3:
            chrome_kwargs['executable_path'] = chromedriver_path
        else:
            # executable_path is deprecated in Selenium 4.1.0 - start using Service() instead.
            chrome_kwargs['service'] = Service(chromedriver_path)  # Selenium API

    return chrome_kwargs


def devtool_command(driver, cmd, params=None):
    """
    Send a command to Chrome via the web driver.
    Example:
        result = devtool_command(driver, "Page.printToPDF", pdf_kwargs)
    """

    if params is None:
        params = {}
    resource = f"/session/{driver.session_id}/chromium/send_command_and_get_result"
    url = driver.command_executor._url + resource
    body = json.dumps({'cmd': cmd, 'params': params})
    response = driver.command_executor._request('POST', url, body)
    if 'status' in response:
        # response dict only contains a "status" key if an error occurred.
        # when "status" is present, the "value" will contain the error message.
        raise ChromePdfException(response.get('value'))
    return response.get('value')


def _get_chromesession_temp_dir():
    """
    Return an absolute path to a folder to use for storing Chrome's files while making a PDF.

    This is used to store user and crash data locally so that users don't run into permissions issues.
    linux default path for crash dumps as of Chrome 99+ is /tmp/Crashpad/
    which can cause permission errors on startup if different user created it.
    create a separate subfolder for each user to avoid permissions conflicts.
    """

    try:
        username = getpass.getuser()
    except Exception:
        username = 'default'

    chromesession_dir = os.path.join(os.path.dirname(__file__), 'chromesession')
    user_dir = os.path.join(chromesession_dir, 'users', username)
    return user_dir



import os
import re

import setuptools
from setuptools import find_packages, setup


_GITHUB_MASTER_ROOT = 'https://github.com/imsweb/django-chromepdf/blob/master/'


def get_version():
    """Return the version string stored at chromepdf.__version__"""
    with open(os.path.join(os.path.dirname(__file__), 'chromepdf', '__init__.py'), encoding='utf8') as fp:
        return re.match(r".*__version__ = '(.*?)'", fp.read(), re.S).group(1)  # @UndefinedVariable


def get_long_description():
    """
    Return a long description consisting of the readme's contents.
    The contents are Markdown, so be sure to set the "long_description_content_type" accordingly.
    Convert all relative urls into absolute ones so that they will display correctly on Pypi.
    """
    with open(os.path.join(os.path.dirname(__file__), 'README.md'), encoding='utf8') as fp:
        contents = fp.read()
        # expand relative github urls to absolute. this will match EVERY url in the document.
        contents = re.sub(r'\[(.+?)\]\((.+?)\)', r'[\1](%s\2)' % _GITHUB_MASTER_ROOT, contents)
        # un-expand urls that were already absolute.
        contents = contents.replace(f'{_GITHUB_MASTER_ROOT}https://', 'https://')
        return contents


def find_chromepdf_packages():
    ""
    return find_packages(exclude=['testapp', '*testapp*', 'testapp.*'])


# setuptools 61.0.0 is the first to support pyproject.toml config file.
# See: https://setuptools.pypa.io/en/latest/history.html#v61-0-0
_PYPROJECT_TOML_SUPPORTED = int(setuptools.__version__.split('.')[0]) >= 61

if _PYPROJECT_TOML_SUPPORTED:
    # rely entirely on pyproject.toml if supported, and ignore this file's contents.
    setup_params = {}

else:
    # deprecated behavior: rely on old setup.py for Python 3.6 and any installs that have
    # setuptools < 61 installed.
    # setuptools 61 does not support Python 3.6, but we still do, because we still support
    # Selenium 3. So we must provide this workaround.
    # Support for Python 3.6 and Selenium 3 will end with the release of ChromePDF 2.0.

    setup_params = dict(
        name='django-chromepdf',
        version=get_version(),
        description='A small Python 3 library that uses Selenium and Google Chrome to convert HTML into a PDF.',
        long_description=get_long_description(),
        long_description_content_type='text/markdown',
        author='Andrew Kukwa',
        author_email='kukwaa@imsweb.com',
        url='https://github.com/imsweb/django-chromepdf',
        license='BSD',
        packages=find_chromepdf_packages(),
        include_package_data=True,
        classifiers=[
            'Development Status :: 5 - Production/Stable',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: BSD License',
            'Operating System :: POSIX',
            'Operating System :: Microsoft :: Windows',
            'Operating System :: MacOS :: MacOS X',
            'Operating System :: OS Independent',
            'Topic :: Utilities',
            'Topic :: Software Development :: Libraries',
            'Programming Language :: Python',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'Programming Language :: Python :: 3.10',
            'Programming Language :: Python :: 3.11',
        ],
        python_requires='~=3.6',
        install_requires=[
            'selenium<5,>=3'
        ],
        project_urls={
            "Source": "https://github.com/imsweb/django-chromepdf",
            "Changelog": "https://github.com/imsweb/django-chromepdf/blob/master/CHANGELOG.md",
        },
    )

if __name__ == '__main__':
    # place setup() in conditional and keep params in global dict so unit tests can access them without calling setup()
    setup(**setup_params)
